/*
  Copyright

  Copyright FGA bitart Furch & Graf intermediaservices GbR, Gartenstr. 5, 
  56220 St. Sebastian. All rights reserved. Copyright in this software is 
  owned by FGA bitart, unless otherwise indicated. Permission to use, copy and
  distribute this software is hereby granted, provided that the above copyright
  notice appear in all copies and that both that copyright notice and this 
  permission notice appear. Note that any product, process or technology described
  in this software may be the subject of other Intellectual Property rights
  reserved by FGA bitart and are not licensed hereunder.

  No Warranty 

  The software is provided "as is" without warranty of any kind,
  either express or implied, including, but not limited to, the
  implied warranties of merchantability, fitness for a particular
  purpose, or non-infringement. This software could include
  technical inaccuracies or typographical errors. Changes are
  periodically made to the software; these changes will be
  incorporated in new editions of the software. FGA bitart may make
  improvements and/or changes in this software at any time
  without notice.

  Limitation Of Liability 

  In no event will FGA bitart be liable for direct, indirect, special,
  incidental, economic, cover, or consequential damages arising
  out of the use of or inability to use this software even if
  advised of the possibility of such damages. Some states do not
  allow the exclusion or limitation of implied warranties or
  limitation of liability for incidental or consequential
  damages, so the above limitation or exclusion may not apply to
  you.

  $Log: mysqldb.c,v $
  Revision 1.9  1997/09/29 21:24:39  andrich
  This is the initial release of my README.

  Revision 1.8  1997/09/29 20:34:43  andrich
  Corrected these CVS keywords.

 */

static char *rcsid="$Id: mysqldb.c,v 1.9 1997/09/29 21:24:39 andrich Exp $";
#if __GNUC__ == 2
#define USE(var) static void * use_##var = (&use_##var, (void *) &var)
USE(rcsid);
#endif

#include "Python.h"
#include "mysql.h"

#define ARG_COUNT(X) \
        (((X) == NULL) ? 0 : (PyTuple_Check(X) ? PyTuple_Size(X) : 1))        

static PyObject *mysqldbError;

/* Connection definitions and method implementation */

typedef struct {
  PyObject_HEAD
  MYSQL handle;
  int valid;
  MYSQL_RES *res;
  long arraysize;
} mysqldbConnectionObject;

staticforward PyTypeObject mysqldbConnectionType;

#define is_mysqldbConnectionObject(v) ((v)->ob_type == &mysqldbConnectionType)

static void mysqldbConnectionObject_dealloc(PyObject* self) {
  mysqldbConnectionObject* m = (mysqldbConnectionObject *)self;

  if (m->valid)
  mysql_close(&(m->handle));
  PyMem_DEL(m);
}

static PyMethodDef mysqldbConnectionObject_methods[];

static PyObject* mysqldbConnectionObject_getattr(mysqldbConnectionObject *m,
                         char *attr) {
  return Py_FindMethod(mysqldbConnectionObject_methods, (PyObject *)m, attr);
}

static PyObject* mysqldbConnectionObject_close(mysqldbConnectionObject *self, 
                         PyObject *args) {

  if (!PyArg_ParseTuple(args, "")) {
  PyErr_SetString(mysqldbError, "usage: close()");
  return NULL;
  }

  if (self->res)
  mysql_free_result(self->res);
 
  if (self->valid)
  mysql_close(&(self->handle));

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_execute(mysqldbConnectionObject *self,
                         PyObject *args) {
  char *operation = NULL;
  PyObject *params = NULL;

  if (!PyArg_ParseTuple(args, "s|o", &operation, &params)) {
  PyErr_SetString(mysqldbError, "usage: execute(operation, [params]) "
                                "operation is a string and params a " 
                                "sequence (!string)");
  return NULL;
  }
  
  if (mysql_query(&(self->handle), operation)) {
  PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
  return NULL;
  } else {
  if (self->res)
    mysql_free_result(self->res);
  if (!(self->res=mysql_store_result(&(self->handle))) &&
    mysql_num_fields(&(self->handle))) {
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  }
  }
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* pythonify_row(MYSQL_RES* res, MYSQL_ROW row, uint num) {
  uint i;
  uint *lengths = NULL;
  PyObject *rowtuple, *fieldobj;
  MYSQL_FIELD *fields;
  
  rowtuple = PyTuple_New(num);
  /*
    get an array of field lengths
  */
  lengths = mysql_fetch_lengths(res);
  fields = mysql_fetch_fields(res);
  for (i=0; i<num; i++) {
    if (row[i]) { 
       switch (fields[i].type) {
        case FIELD_TYPE_SHORT:    
        case FIELD_TYPE_INT24:
        case FIELD_TYPE_LONG: 
          fieldobj = PyInt_FromLong(atol(row[i]));   
          break;   
        case FIELD_TYPE_LONGLONG:
          fieldobj = PyLong_FromLong(atol(row[i]));
          break;
        case FIELD_TYPE_CHAR:    
        case FIELD_TYPE_STRING:
        case FIELD_TYPE_VAR_STRING:   
        case FIELD_TYPE_DATE:   
        case FIELD_TYPE_TIME:   
        case FIELD_TYPE_DATETIME:   
        case FIELD_TYPE_TIMESTAMP:   
          fieldobj = PyString_FromString(row[i]);
          break;  
        case FIELD_TYPE_TINY_BLOB:   
        case FIELD_TYPE_MEDIUM_BLOB:   
        case FIELD_TYPE_LONG_BLOB:   
        case FIELD_TYPE_BLOB:   
          fieldobj = PyString_FromStringAndSize(row[i], lengths[i]);   
          break;   
        case FIELD_TYPE_DECIMAL:
        case FIELD_TYPE_FLOAT:
        case FIELD_TYPE_DOUBLE:    
          fieldobj = PyFloat_FromDouble(atof(row[i]));   
          break;   
        default:    
          fieldobj = NULL;   
          break;   
      }
    } else {
      Py_INCREF(Py_None);
      fieldobj = Py_None;
    }
    if (fieldobj == NULL) { 
      Py_DECREF(rowtuple); 
      PyErr_SetString(mysqldbError,"mySQL database returned bad value"); 
      return NULL; 
    } 
     PyTuple_SetItem(rowtuple,i,fieldobj);
   }
   return rowtuple;
}

static PyObject* mysqldbConnectionObject_fetchone(mysqldbConnectionObject *self,
                          PyObject *args) {
  MYSQL_ROW row;
  uint num;

  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: fetchone()");
    return NULL;
  }

  if ((row = mysql_fetch_row(self->res))) {
    num = mysql_num_fields(self->res);
    return pythonify_row(self->res, row, num);    
  }
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_fetchall(mysqldbConnectionObject *self,             
                                                  PyObject *args) {
  MYSQL_ROW row;
  PyObject *reslist, *restuple;
  uint num;
  
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: fetchall()");
    return NULL;
  }

  reslist = PyList_New(0);
  num = mysql_num_fields(self->res);
  
  while (row=mysql_fetch_row(self->res)) {
    restuple = pythonify_row(self->res, row, num);
    PyList_Append(reslist,restuple);
    Py_DECREF(restuple);
  }

  return reslist;
}

static PyObject* mysqldbConnectionObject_fetchmany(mysqldbConnectionObject *self,
                                                   PyObject *args) {
  int size, i;
  uint num;
  MYSQL_ROW row;
  PyObject *reslist, *restuple;

  if (!PyArg_ParseTuple(args, "|i", &size)) {
    PyErr_SetString(mysqldbError, "usage: fetchmany([size]); size is an int");
    return NULL;
  }

  reslist = PyList_New(0);
  num = mysql_num_fields(self->res);

  for (i=0; i<self->arraysize; i++) {
    if (row=mysql_fetch_row(self->res)) {
      restuple = pythonify_row(self->res, row, num);
      PyList_Append(reslist, restuple);
      Py_DECREF(restuple);
    } else {
      break;
    }
  }

  return reslist;
}

static PyObject* mysqldbConnectionObject_createdb(mysqldbConnectionObject *self,
                                                  PyObject *args) {
  char *dbname;

  if (!PyArg_ParseTuple(args, "s", &dbname)) {
    PyErr_SetString(mysqldbError, "usage: createdb(dbname); dbname must be a" \
                                  "string");
    return NULL;
  }
  if (mysql_create_db(&(self->handle),dbname)) {
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_dropdb(mysqldbConnectionObject *self,
                                                PyObject *args) {
  char *dbname;

  if (!PyArg_ParseTuple(args, "s", &dbname)) {
    PyErr_SetString(mysqldbError, "usage: dropdb(dbname); dbname must be a" \
                                  "string");
    return NULL;
  }
  if (mysql_drop_db(&(self->handle),dbname)) {
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_clientinfo(mysqldbConnectionObject *self,
                                                    PyObject *args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: clientinfo()");
    return NULL;
  }
  return PyString_FromString(mysql_get_client_info());
}

static PyObject* mysqldbConnectionObject_hostinfo(mysqldbConnectionObject *self,
                                                    PyObject *args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: hostinfo()");
    return NULL;
  }
  return PyString_FromString(mysql_get_host_info(&(self->handle)));
}

static PyObject* mysqldbConnectionObject_protoinfo(mysqldbConnectionObject *self,
                                                   PyObject *args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: protoinfo()");
    return NULL;
  }
  return PyInt_FromLong((long)mysql_get_proto_info(&(self->handle)));
}

static PyObject* mysqldbConnectionObject_serverinfo(mysqldbConnectionObject *self,
                                                    PyObject *args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: serverinfo()");
    return NULL;
  }
  return PyString_FromString(mysql_get_server_info(&(self->handle)));
}

static PyObject* mysqldbConnectionObject_reload(mysqldbConnectionObject *self,
                                                PyObject *args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: reload()");
    return NULL;
  }

  if (mysql_reload(&(self->handle))) {
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  }
  
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_selectdb(mysqldbConnectionObject *self,
                                                  PyObject *args) {
  char *dbname;

  if (!PyArg_ParseTuple(args,"s", &dbname)) {
    PyErr_SetString(mysqldbError, "usage: selectdb(dbname); dbname must be a" \
                                  "string");
    return NULL;
  }

  if (mysql_select_db(&(self->handle), dbname)) {
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_shutdown(mysqldbConnectionObject *self,
                                                  PyObject *args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: shutdown()");
    return NULL;
  }
  if (mysql_shutdown(&(self->handle))) {
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_stat(mysqldbConnectionObject *self,
                                              PyObject *args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: stat()");
    return NULL;
  }

  return PyString_FromString(mysql_stat(&(self->handle)));
}

static PyObject* mysqldbConnectionObject_listdbs(mysqldbConnectionObject *self,
                                                  PyObject *args) {
  char *wild = NULL;
  MYSQL_RES* temp;

  if (!PyArg_ParseTuple(args, "|s", &wild)) {
    PyErr_SetString(mysqldbError, "usage: listdbs([wild]) "
                                  "wild is a sql regex string (as in like)");
    return NULL;
  }
 
  temp = mysql_list_dbs(&(self->handle), wild);
  
  if (!temp) {
    fprintf(stderr, "ERROR");
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  } 
  if (self->res)
    mysql_free_result(self->res);
  self->res = temp;
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_listtables(mysqldbConnectionObject *self,
                                                     PyObject *args) {
  char *wild = NULL;
  MYSQL_RES* temp;

  if (!PyArg_ParseTuple(args, "|s", &wild)) {
    PyErr_SetString(mysqldbError, "usage: listtables([wild]); "
                                  "wild is a sql regex string (as in like)");
    return NULL;
  }
 
  temp = mysql_list_tables(&(self->handle), wild);
  
  if (!temp) {
    fprintf(stderr, "ERROR");
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  } 
  if (self->res)
    mysql_free_result(self->res);
  self->res = temp;
  Py_INCREF(Py_None);
  return Py_None;
}

/* 
  This function does work yet, seems to be a problem with the mysql API call
*/
static PyObject* mysqldbConnectionObject_listfields(mysqldbConnectionObject *self,
                                                     PyObject *args) {
  char *table; 
  char *wild = "%";
  MYSQL_RES* temp;

  if (!PyArg_ParseTuple(args, "s|s", &table, &wild)) {
    PyErr_SetString(mysqldbError, "usage: listfields(table, [wild]); "
                                  "table is a string, "
                                  "wild is a sql regex string (as in like)");
    return NULL;
  }
 
  temp = mysql_list_fields(&(self->handle), table, wild);
  
  if (!temp) {
    fprintf(stderr, "ERROR");
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  } 
  if (self->res)
    mysql_free_result(self->res);
  self->res = temp;
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_listprocesses(mysqldbConnectionObject *self,
                                                        PyObject *args) {
  MYSQL_RES* temp;

  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: listprocesses()");
    return NULL;
  }
 
  temp = mysql_list_processes(&(self->handle));
  
  if (!temp) {
    fprintf(stderr, "ERROR");
    PyErr_SetString(mysqldbError, mysql_error(&(self->handle)));
    return NULL;
  } 
  if (self->res)
    mysql_free_result(self->res);
  self->res = temp;
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* mysqldbConnectionObject_affectedrows(mysqldbConnectionObject *self,
                                                      PyObject* args) {
  if (!PyArg_ParseTuple(args, "")) {
    PyErr_SetString(mysqldbError, "usage: affectedrows()");
    return NULL;
  }

  return PyInt_FromLong((long)mysql_affected_rows(&(self->handle)));
}

static PyObject* mysqldbConnectionObject_arraysize(mysqldbConnectionObject *self,
                                                   PyObject* args) {
  long size;
  
  if (!PyArg_ParseTuple(args, "|i", &size)) {
    PyErr_SetString(mysqldbError, "usage: arraysize([size]); size is in an"
                                  "int");
    return NULL;
  }

  if (ARG_COUNT(args)) {
    if (size <= 0) {
      self->arraysize = 1;
    } else {
      self->arraysize = size;
    }
  } else {
    return PyInt_FromLong(self->arraysize);
  }
  
  Py_INCREF(Py_None);
  return Py_None;
}

static PyMethodDef mysqldbConnectionObject_methods [] = {
  /* API Methods */
  {"arraysize", (PyCFunction)mysqldbConnectionObject_arraysize, 1},
  {"close", (PyCFunction)mysqldbConnectionObject_close, 1},
  {"execute", (PyCFunction)mysqldbConnectionObject_execute, 1},
  {"fetchone", (PyCFunction)mysqldbConnectionObject_fetchone, 1},
  {"fetchmany", (PyCFunction)mysqldbConnectionObject_fetchmany, 1},
  {"fetchall", (PyCFunction)mysqldbConnectionObject_fetchall, 1},
  /* NON API Methods */
  {"affectedrows", (PyCFunction)mysqldbConnectionObject_affectedrows, 1},
  {"createdb", (PyCFunction)mysqldbConnectionObject_createdb, 1},
  {"dropdb", (PyCFunction)mysqldbConnectionObject_dropdb, 1},
  {"clientinfo", (PyCFunction)mysqldbConnectionObject_clientinfo, 1},
  {"hostinfo", (PyCFunction)mysqldbConnectionObject_hostinfo, 1},
  {"protoinfo", (PyCFunction)mysqldbConnectionObject_protoinfo, 1},
  {"serverinfo", (PyCFunction)mysqldbConnectionObject_serverinfo, 1},
  {"listdbs", (PyCFunction)mysqldbConnectionObject_listdbs, 1},
  {"listtables", (PyCFunction)mysqldbConnectionObject_listtables, 1},
  /*
    {"listfields", (PyCFunction)mysqldbConnectionObject_listfields, 1},
  */
  {"listprocesses", (PyCFunction)mysqldbConnectionObject_listprocesses, 1},
  {"reload", (PyCFunction)mysqldbConnectionObject_reload, 1},
  {"selectdb", (PyCFunction)mysqldbConnectionObject_selectdb, 1},
  {"shutdown", (PyCFunction)mysqldbConnectionObject_shutdown, 1},
  {"stat", (PyCFunction)mysqldbConnectionObject_stat, 1},
  {NULL, NULL}                                        /* sentinel */
};

static char mysqldbConnectionType__doc__[] = 
"A connection to the mysql DB Server."
;

static PyTypeObject mysqldbConnectionType = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,                                            /* ob_size */
  "mysql",                                      /* tp_name */
  sizeof(mysqldbConnectionObject),              /* tp_basicsize */          
  0,                                            /* tp_itemsize */ 
  /* methods */
  (destructor)mysqldbConnectionObject_dealloc,  /* tp_dealloc */
  (printfunc)0,                                 /* tp_print */
  (getattrfunc)mysqldbConnectionObject_getattr, /* tp_getattr */
  (setattrfunc)0,                               /* tp_setattr */
  (cmpfunc)0,                                   /* tp_compare */
  (reprfunc)0,                                  /* tp_repr */
  0,                                            /* tp_as_number */
  0,                                            /* tp_as_sequence */
  0,                                            /* tp_as_mapping */
  (hashfunc)0,                                  /* tp_hash */
  (ternaryfunc)0,                               /* tp_call */
  (reprfunc)0,                                  /* tp_str */
  0L,0L,0L,0L,
  mysqldbConnectionType__doc__                  /* Doc String */
};      

/* Module definition and module methods */

static PyObject* mysqldb_connect(PyObject *self, PyObject *args, PyObject *kws) {
  char *dbhost = NULL, *dbuser = NULL, *dbpass = NULL, *dbname = NULL;
  char *kwd[] = {"dbhost", "dbuser", "dbpasswd", "dbname", NULL};
  mysqldbConnectionObject *newConnection;
  MYSQL newHandle;
  
  if (!PyArg_ParseTupleAndKeywords(args, kws, "|ssss", kwd, 
                   &dbhost, &dbuser, &dbpass, &dbname)) {
  PyErr_SetString(mysqldbError, 
          "this method supports up to 4 optional string arguments" \
          "or keyword arguments: dbhost, dbuser, dbpasswd, dbname");
  return NULL;
  }
  
  if (!mysql_connect(&newHandle, dbhost, dbuser, dbpass)) {
  PyErr_SetString(mysqldbError, mysql_error(&newHandle));
  return NULL;
  }

  if (dbname) {
  if (mysql_select_db(&newHandle, (const char*)dbname)) {
    PyErr_SetString(mysqldbError, mysql_error(&newHandle));
    return NULL;
  }
  }

  newConnection = PyObject_NEW(mysqldbConnectionObject, &mysqldbConnectionType);
  if (!newConnection) {
  PyErr_SetString(mysqldbError, "Unable to create new connection object");
  return NULL;
  }

  newConnection->valid = 1;
  newConnection->arraysize = 1;
  newConnection->handle = newHandle;
  newConnection->res = NULL;
  
  return ((PyObject *) newConnection);
}

static struct PyMethodDef mysqldb_methods[] = {
  {"mysqldb", (PyCFunction)mysqldb_connect, 3},
  {NULL,NULL}
};

/* Initialization function for the module (*must* be called initmysqldb) */

static char mysqldb__doc__[] = 
"Python interface to mysql DB"
;

void initmysqldb() {
  PyObject *module, *dict;
  PyObject *mysqldbString, *mysqldbRaw, *mysqldbNumber, *mysqldbDate, *mysqldbRowid;

  module = Py_InitModule4("mysqldb", 
              mysqldb_methods,
              mysqldb__doc__,
              (PyObject*)NULL,
              PYTHON_API_VERSION);

  dict = PyModule_GetDict(module);

  mysqldbError = PyString_FromString("mysqldb.error");
  
  PyDict_SetItemString(dict, "error", mysqldbError);

  /* final error checking */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module mysqldb");
}
